! ESA (C) 2000-2021
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: esensmap_mod
!
! Authors:  Hermann Brunner - AIP
! Updates:  Iris Traulsen - AIP
!
! Description: calculate global EPIC sensitivity map
!
! Comments:
!
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!-

module esensmap_mod

  use dal
  use errorHandling
  use param
  use edetect_io_sens
  use esens_deltac_mod

  implicit none

  private

  public esensmap

contains

  !---------------------------------------------------------------------
  subroutine esensmap
  !---------------------------------------------------------------------

    real :: ul, back,texp
    real, dimension(max_ins) :: xcen, ycen
    character(len=1024) :: ostring, fname
!    character(len=64), dimension(max_ima) :: image
    character(len=16), dimension(max_ima) :: instrume
    character(len=1024), dimension(max_ima) :: c_bgima, c_expima, c_mask
    character(len=10) :: stat, outunit
    integer :: i, j, k, k0, k1, ix, iy, ix1, iy1, inst, ipoint
    integer :: nout, necf
    real, dimension(3) :: xoff, yoff
    integer, dimension(max_ima) :: band
    real(kind=single), dimension(:,:), pointer :: sensrate, sensflux
    real(kind=single), parameter :: psfrac = 0.3

    real :: gammp

    real :: doff
    real(kind=double) :: fluxconv

    integer :: nbands                          !! number of energy bands
    integer :: ntheta                          !! number of off-axis angles
    integer :: iinst, iband                    !! loop variables
    integer, dimension(2) :: coords            !! pair of coordinates

    integer, dimension(:,:), allocatable :: camera

    real(kind=double), dimension(max_ins) :: equdeltac   !! DeltaC calculation
    real(kind=double)                     :: p, q, dof, x, bound
    integer                               :: status, which

    integer :: dtvalues(1:8), psf_id

    character(124) :: psfmodel

    type(DataSetT) :: set,setO, set1, setO1
    type(ArrayT) :: arr, arr1


! input files

!   print *, "get bg_image"
    call get_files("bkgimagesets",c_bgima, instrume, band,.true.)
!   print *, "get expo_image"
    call get_files("expimagesets",c_expima, instrume, band,.false.)
!   print *, "get det_mask"
    call get_files("detmasksets",c_mask, instrume, band,.false.)

! read general parameters
    ul = realParameter("mlmin")
    stat = stringParameter("statistic")       ! validity of input tested by sas
    outunit = stringParameter("outunit")
    nout=parameterCount("sensimageset")
    if ((nout.gt.1 .and. outunit.ne."both") .or. nout.gt.2) &
         call warning("FileMismatch", "More than one file name provided while &
                  &one output file requested. Using the first file name only.")
    if (nout.eq.1 .and. outunit.eq."both") &
         call error("FileMismatch", "outunit=both requires two output file &
              &names, first one for rate units and second one for flux units.")
    necf=parameterCount("ecf")
    allocate(ecf(necf))
    ecf=1.0
    do i=1, necf
      ecf(i) = realParameter("ecf",i-1)
    enddo

! coordinates
    xoff=0.
    yoff=0.
    inst=1

    crpix1=-2d30
    crpix2=-2d30
    crval1=-2d30
    crval2=-2d30
    crdelt1=-2d30
    crdelt2=-2d30

! read files
    do i = 1,num_instr
      do j = 1,num_image(i)
        ipoint=ip(i,j)

        fname = c_bgima(ip(i,j))
        write(errstr,*) "Opening file ",fname
        call message(AppMsg,Verbose)
        call co_bgima(fname,npix,crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                      xoff,yoff,inst,ipoint,nim,stat)

        fname = c_expima(ip(i,j))
        write(errstr,*) "Opening file ",fname
        call message(AppMsg,Verbose)
        call co_expima(fname,npix,crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                       xoff,yoff,inst,ipoint,nim)
      end do

      ipoint=ip(i,1)
      fname = c_mask(i)
      write(errstr,*) "Opening file ",fname
      call message(AppMsg,Verbose)
      call co_mask(fname,npix,crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                   xoff,yoff,inst,ipoint,i)

      do j = 2,num_image(i)
        do iy = imymin,imymax
          do ix = imxmin,imxmax
            mask_image(ix,iy,ip(i,j))=mask_image(ix,iy,ipoint)
          end do
        end do
      end do

    end do
    call message(AppMsg,Verbose,"Data read in. Calculating the sensitivity ...")

    if (crpix1<-1d30) crpix1=0d0
    if (crpix2<-1d30) crpix2=0d0
    if (crval1<-1d30) crval1=0d0
    if (crval2<-1d30) crval2=0d0
    if (crdelt1<-1d30) crdelt1=0d0
    if (crdelt2<-1d30) crdelt2=0d0

! open (first) output file
    ostring=stringParameter("sensimageset",0)
    setO=dataset(ostring,Create)
    set=dataset(c_bgima(1),Read)

    call cp_head(set,setO,"EPIC EXPOSURE SENSITIVITY MAP", 2)

    call setAttribute(setO, "HDUCLASS", "OGIP    ", &
                            "Format conforms to OGIP/GSFC conventions")
    call setAttribute(setO, "HDUCLAS1", "IMAGE   ", &
                            "File contains an image")
    if (outunit.ne."flux") then
       call setAttribute(setO, "HDUCLAS2", "GENERIC", &
                               "Sensitivity map in rate units")
    else
       call setAttribute(setO, "HDUCLAS2", "GENERIC", &
                               "Sensitivity map in flux units")
    endif

    arr=addArray(setO,"PRIMARY",Real32,npix)
    sensrate => real32Array2Data(arr)
    sensrate = -1.0


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! POISSON mode

    if (stat == "poisson") then

       ! prepare flux conversion
       if (outunit.ne."rate") then
          if ((nim>1) .and. (necf>1)) &     ! poisson mode: one ecf
               call warning("SingleECF", &
                            "Images were summed. Only the first ECF is used.")
          fluxconv = 1.d-11 / ecf(1)
       endif

       ! loop over image pixels
       do iy = 1,npix(2)
          do ix = 1,npix(1)

             back = 0.0
             texp = 0.0

             ! sum images, if several are provided
             do ipoint = 1,nim

                do iy1 = iy-1,iy+1
                   do ix1 = ix-1,ix+1
                      if (ix1 >= 1 .and. ix1 <= npix(1) .and. &
                           iy1 >= 1 .and. iy1 <= npix(2)) then
                         if (mask_image(ix1,iy1,ipoint)) then
                            back = back + bg_image(ix1,iy1,ipoint)
                            texp = texp + exp_image(ix1,iy1,ipoint)
                         end if
                      end if
                   end do
                end do

             end do

             ! calculate likelihood and limiting rate
             if (texp > 0.0 .and. back > 0.0) then

                k1=1
                do while (gammp(real(k1),back) <= ul)
                   k1=k1*2
                end do
                k0=k1/2
                do while (k1>k0+1)
                   k=(k0+k1)/2
                   if (gammp(real(k),back) > ul) then
                      k1=k
                   else
                      k0=k
                   end if
                end do

                sensrate(ix-1,iy-1) = 9.0*nim/psfrac * (real(k1)-back)/texp
             end if

          end do
       end do

    endif


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! DELTAC mode

    ! DeltaC sensitivity routine based on code by Christian Motch (OAS):
    ! Calculate limiting full-band rates for a given (minimum) detection
    ! likelihood using the ellbeta PSF and DeltaC statistic

    if (stat == "deltac") then

       ! read and check deltac-specific parameters
       if (parameterCount("energies").lt.nim) &
            call error("FileMismatch", "Number of mid-band energies has to &
            &identical with the number of images")
       allocate(energies(nim))
       do i=1,nim
          energies(i) = realParameter("energies",i-1)
       end do
       if (necf.lt.nim) call warning("FileMismatch", &
            "Number of ECFs not equal number of images. ECFs default to 1.0.")
       psfmodel = stringParameter("psfmodel")
       select case (psfmodel)
       case ("slew")
          psf_id = 4
       case default  !! ellbeta
          psf_id = 3
       end select
       doff = realParameter("doff")
       ecut = intParameter("ecut")
       isize = 2*ecut+1            ! passed to esens_deltac_mod

       ! normalize ECF count rates to 1cnt/s in all cameras
       fluxconv = sum(ecf(1:nim))
       ecf = ecf / fluxconv
       ! corresponding flux (ECFs are in units of cnts/s / (1e-11erg/cm2/s) )
       fluxconv = 1d-11 / fluxconv

       ! number of active energy bands
       nbands = maxval(num_image)

       ! number of active cameras in each pixel, derived from mask_image
       allocate(camera(npix(1),npix(2)))
       camera = 0
       do iinst = 1,num_instr
          where(mask_image(:,:,ip(iinst,1)))  camera = camera+1
       end do

       ! calculate the limiting DeltaC corresponding to the minimum likelihood
       q = exp(-ul)   	                       ! reminder: ul = mlmin
       p = 1.0D+00 - q
       which = 2                               ! Calculate X from P, Q and DF
       do iinst = 1,num_instr
          ! only point-like sources considered here
          dof = 2.0 + dfloat(nbands*iinst)
          call cdfchi (which, p, q, x, dof, status, bound)        !! from cdflib
          equdeltac(iinst) = x
          write(errstr, '(a,i4,a,f8.4)') &
               " Limiting DeltaC for ", int(dof), " degrees of freedom: ", x
          call message(AppMsg, Verbose)
       enddo

       ! read PSFs into an array psfima(ecut*2+1,ecut*2+1,nim,ntheta)
       call rpsf(psf_id, doff, ntheta)

       ! check DeltaC at the pixel with maximum exposure, summing up the images
       coords = maxloc(sum(exp_image(:,:,:), 3, mask_image(:,:,:)))
       dcd_comp = equdeltac(camera(coords(1),coords(2)))
       dcd_x = coords(1)
       dcd_y = coords(2)
       dcd_xmin = max(  1,        1+ecut-dcd_x)
       dcd_xmax = min(isize,npix(1)+ecut-dcd_x)
       dcd_ymin = max(  1,        1+ecut-dcd_y)
       dcd_ymax = min(isize,npix(2)+ecut-dcd_y)

       if (deltac_diff(0d0) .gt. 0)     call error("InvalidStat", &
            "Issue fitting Deltac statistic: deltac_diff(0.0) > 0")

       ! fit
       if (verbosity(appMsg) >= Verbose) then    !!! if verbose, inform on start
          call date_and_time(values=dtvalues)
          write(errstr, '(a,i5,a,i5,a,"  -  ",i4,"-",i2.2,"-",i2.2," ",&
               &i2.2,":",i2.2,":",i2.2,".",i3.3)') &
               " Sensitivity in ",imusex," by ",imusey, " exposed pixels", &
               dtvalues(1:2), &
               dtvalues(3)+(isign(1,dtvalues(5)-dtvalues(4)/60)-1)/2, &
               mod(dtvalues(5)-dtvalues(4)/60 + 24,24), dtvalues(6:8)
          call message(AppMsg,Verbose)
       endif

       iy1 = 1
       do iy = imymin,imymax
          ix1 = 1
          do ix = imxmin,imxmax
             ! valid images only
             if (camera(ix,iy).gt.0) then
                ! sensitivity value, using closest PSF of one instrument
                sensrate(ix-1,iy-1) = &
                     deltac_fit(equdeltac(camera(ix,iy)), ix, iy, nbands) &
                     / fluxed_stot(maxval(ipsf1(ix1,iy1,1:nim)))
             endif
             ix1 = ix1+1
          enddo
          ! if verbose, inform on runtime
          if (verbosity(appMsg) >= Verbose) then
             if (mod(iy1,50).eq.0 .or. iy1.eq.npix(2)) then
                call date_and_time(values=dtvalues)
                write(errstr, '(i21,a,"  -  ",i4,"-",i2.2,"-",i2.2," ",&
                      &i2.2,":",i2.2,":",i2.2,".",i3.3)') &
                      iy1, " lines processed", &
                      dtvalues(1:2), &
                      dtvalues(3)+(isign(1,dtvalues(5)-dtvalues(4)/60)-1)/2, &
                      mod(dtvalues(5)-dtvalues(4)/60 + 24,24), dtvalues(6:8)
                call message(AppMsg,Verbose)
             endif
          endif
          iy1 = iy1+1
       enddo

    endif

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    ! output of sensitivity images

    ! convert to flux?
    if (outunit.ne."rate") then

       ! open second output file?
       if (outunit.eq."both") then
          ! open second file
          ostring=stringParameter("sensimageset",1)
          setO1=dataset(ostring,Create)
          set1=dataset(c_bgima(1),Read)
          call cp_head(set1,setO1,"EPIC EXPOSURE SENSITIVITY MAP", 2)
          call setAttribute(setO1, "HDUCLASS", "OGIP    ", &
               "Format conforms to OGIP/GSFC conventions")
          call setAttribute(setO1, "HDUCLAS1", "IMAGE   ", &
               "File contains an image")
          call setAttribute(setO1, "HDUCLAS2", "GENERIC", &
               "Sensitivity map in flux units")
          arr1=addArray(setO1,"PRIMARY",Real32,npix)
          ! set flux pointer to second output file
          sensflux => real32Array2Data(arr1)
          sensflux = -1e-10
       else
          ! set flux pointer to single output file
          sensflux => real32Array2Data(arr)
       endif

       ! flux conversion
       sensflux = sensrate * fluxconv
       ! set masked areas to negative value
       where(sensrate < 0) sensflux=-1e-10

       if (outunit.eq."both") then
          ! close and write second file
          call release(arr1)
          call release(set1)
          call release(setO1)
       endif

    endif    ! end of flux conversion

    call release(arr)
    call release(set)
    call release(setO)

    call message(AppMsg,Verbose,"Done")

  end subroutine esensmap

end module esensmap_mod
